# A hook that measures the amount of time spent in various events.

import logging
import time
import typing as t
from collections import defaultdict


class TimerHook:

    # Relatively high prehook/posthook priority so that TimerHook gets
    # called soon before and after a wrapped function.
    prehook_priority: int = 8
    posthook_priority: int = 8

    event_times: t.DefaultDict[str, float]

    def __init__(self) -> None:
        self.event_times = defaultdict(float)

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> float:
        # Return the current time so that it can be used by posthook()
        return time.time()

    def posthook(self, event_name: str, result: t.Any, context: float) -> None:
        # The context stores the time when the prehook was called. We can calculate the
        # total time spent in the event as roughly (not accounting for other hooks) equal
        # to the current time minus the time returned by prehook().
        current_time = time.time()
        self.event_times[event_name] += current_time - context

    def reset(self) -> None:
        self.event_times.clear()

    def log_results(self, logger: logging.Logger) -> None:
        logger.info("%s results:", self.__class__.__name__)
        for event in sorted(self.event_times):
            logger.info("    Time spent in %s: %f", event, self.event_times[event])



# Testing utilities and base classes for testing Seagrass

import logging
import typing as t
from io import StringIO
from seagrass import Auditor
from seagrass.base import ProtoHook


class SeagrassTestCaseMixin:

    logging_output: StringIO
    logger: logging.Logger
    auditor: Auditor

    def setUp(self) -> None:
        # Set up an auditor with a basic logging configuration
        self.logging_output = StringIO()
        fh = logging.StreamHandler(self.logging_output)
        fh.setLevel(logging.DEBUG)

        formatter = logging.Formatter("(%(levelname)s) %(message)s")
        fh.setFormatter(formatter)

        self.logger = logging.getLogger("test.seagrass")
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(fh)

        # Create a new auditor instance with the logger we just
        # set up
        self.auditor = Auditor(logger=self.logger)


class HookTestCaseMixin(SeagrassTestCaseMixin):
    """A base testing class for auditor hooks."""

    hook: ProtoHook
    hook_gen: t.Optional[t.Callable[[], ProtoHook]] = None

    # A list of all of the interfaces that the hook is expected
    # to satisfy.
    check_interfaces: t.Optional[t.Tuple[t.Type, ...]] = None

    @property
    def hook_name(self) -> str:
        return self.hook.__class__.__name__

    def setUp(self):
        super().setUp()

        if self.hook_gen is not None:
            self.hook = self.hook_gen()

    def test_hook_satisfies_interfaces(self):
        CheckableProtoHook = t.runtime_checkable(ProtoHook)
        self.assertIsInstance(
            self.hook,
            CheckableProtoHook,
            f"{self.hook_name} does not satisfy the ProtoHook interface",
        )

        interfaces = [] if self.check_interfaces is None else self.check_interfaces

        for interface in interfaces:
            self.assertIsInstance(
                self.hook,
                interface,
                f"{self.hook_name} does not satisfy the {interface.__name__} interface",
            )



import logging
import sys
import typing as t
import warnings
from collections import defaultdict


class FileOpenInfo(t.NamedTuple):
    filename: str
    mode: str
    flags: int


class FileOpenHook:
    """An event hook for tracking calls to the Python standard
    library's `open` function."""

    # Give this hook slightly higher priority by default so that
    # we can avoid counting calls to open that occur in other
    # hooks.
    prehook_priority: int = 3
    posthook_priority: int = 3

    file_open_counter: t.DefaultDict[str, t.Counter[FileOpenInfo]]
    track_nested_opens: bool
    __current_event_stack: t.List[str]

    def __init__(self, track_nested_opens: bool = False) -> None:
        self.file_open_counter = defaultdict(t.Counter[FileOpenInfo])
        self.track_nested_opens = track_nested_opens
        self.__current_event_stack = []

        # Add the __sys_audit_hook closure as a new audit hook
        sys.addaudithook(self.__sys_audit_hook)

    def __sys_audit_hook(self, event: str, args: t.Tuple[t.Any, ...]) -> None:
        try:
            if len(self.__current_event_stack) > 0 and event == "open":
                filename, mode, flags = args
                info = FileOpenInfo(filename, mode, flags)

                # When track_nested_opens is set, we increment the number of opens
                # for every (unique) event in the __current_event_stack. Otherwise,
                # we only count it for the most recent event.
                if self.track_nested_opens:
                    for event in set(self.__current_event_stack):
                        self.file_open_counter[event][info] += 1
                else:
                    event = self.__current_event_stack[-1]
                    self.file_open_counter[event][info] += 1

        except Exception as ex:
            # In theory we shouldn't reach this point, but if we don't include
            # this try-catch block then we could hit an infinite loop if an
            # error *does* occur.
            warnings.warn(
                f"{ex.__class__.__name__} raised while calling {self.__class__.__name__}'s audit hook: {ex}"
            )

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> None:
        self.__current_event_stack.append(event_name)

    def posthook(
        self,
        event_name: str,
        result: t.Any,
        context: None,
    ) -> None:
        # Do nothing -- we defer fixing the __current_event_stack to the cleanup stage
        pass

    def cleanup(self, event_name: str, context: None) -> None:
        self.__current_event_stack.pop()

    def reset(self) -> None:
        self.file_open_counter.clear()

    def log_results(self, logger: logging.Logger) -> None:
        logger.info("%s results (file opened, count):", self.__class__.__name__)
        for event in sorted(self.file_open_counter):
            logger.info("  event %s:", event)
            for (info, count) in self.file_open_counter[event].items():
                logger.info(
                    "    %s (mode=%s, flags=%s): opened %d times",
                    info.filename,
                    info.mode,
                    hex(info.flags),
                    count,
                )



import logging
import pstats
import typing as t
from io import StringIO
from cProfile import Profile

# Type alias to represent the input types that are allowed as "restrictions"
R = t.Union[str, int, float]


class ProfilerHook:
    """A Seagrass hook that uses the built-in cProfile module to collect and log performance
    statistics on events."""

    profile: Profile
    sort_keys: t.Union[t.Tuple[int], t.Tuple[str, ...]]
    restrictions: t.Tuple[R, ...]

    # Set a high prehook_priority and posthook_priority to ensure
    # that the profiler only gets called directly before and after
    # the event.
    prehook_priority: int = 10
    posthook_priority: int = 10

    def __init__(
        self,
        sort_keys: t.Union[t.Union[str, int], t.Tuple[str, ...]] = tuple(),
        restrictions: t.Union[R, t.Tuple[R, ...]] = tuple(),
    ) -> None:
        """Create a new ProfilerHook.

        :param Union[K,Tuple[K,...]] sort_keys: a key or list of keys to use to sort the output
            generated by :py:meth:`log_results`. The available keys and their meanings are the same
            as those of `pstats.Stats.sort_stats`_.
        :param Union[R,Tuple[R,...]] restrictions: a restriction or list of restrictions to use
            for the output generated by :py:meth:`log_results`. The available restrictions and
            their meanings are the same as those of `pstats.Stats.print_stats`_.

        .. _pstats.Stats.sort_stats: https://docs.python.org/3/library/profile.html#pstats.Stats.sort_stats
        .. _pstats.Stats.print_stats: https://docs.python.org/3/library/profile.html#pstats.Stats.print_stats
        """

        if isinstance(sort_keys, str):
            self.sort_keys = (sort_keys,)
        elif isinstance(sort_keys, int):
            self.sort_keys = (sort_keys,)
        else:
            self.sort_keys = sort_keys

        if isinstance(restrictions, (str, int, float)):
            self.restrictions = (restrictions,)
        else:
            self.restrictions = restrictions

        self.profile = Profile()

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> None:
        # Start profiling
        self.profile.enable()

    def posthook(self, event_name: str, result: t.Any, context: None) -> None:
        # Do nothing -- we defer disabling profiling to the cleanup stage
        pass

    def cleanup(self, event_name: str, context: None) -> None:
        # Stop profiling
        self.profile.disable()

    def get_stats(self, **kwargs) -> t.Optional[pstats.Stats]:
        """Return the profiling statistics as a pstats.Stats class.

        :param kwargs: Keyword arguments to pass to the constructor of `pstats.Stats`_.
        :return: an instance of `pstats.Stats`_. If no profiling information was collected,
            this function will return ``None`` instead.
        :rtype: Optional[pstats.Stats]

        .. _pstats.Stats: https://docs.python.org/3/library/profile.html#pstats.Stats
        """
        if self.profile.getstats() == []:  # type: ignore
            return None
        else:
            return pstats.Stats(self.profile, **kwargs)

    def reset(self) -> None:
        """Reset the internal profiler."""
        self.profile.clear()  # type: ignore

    def log_results(self, logger: logging.Logger) -> None:
        """Log the results captured by ProfilerHook."""
        logger.info("Results from %s:", self.__class__.__name__)

        # Dump results to an in-memory stream
        output = StringIO()
        if (stats := self.get_stats(stream=output)) is None:
            logger.info("   (no samples were collected)")
            return

        stats.sort_stats(*self.sort_keys).print_stats(*self.restrictions)

        # Now take results from the in-memory stream and log them using the provided logger.
        logger.info("")
        output.seek(0)
        for line in output.readlines():
            logger.info("    " + line.rstrip())



# Tests for the TimerHook auditing hook

import time
import unittest
from test.utils import HookTestCaseMixin
from seagrass.base import LogResultsHook, ResettableHook
from seagrass.hooks import TimerHook


class TimerHookTestCase(HookTestCaseMixin, unittest.TestCase):

    hook_gen = TimerHook
    check_interfaces = (LogResultsHook, ResettableHook)

    def test_hook_function(self):
        ausleep = self.auditor.audit("test.time.sleep", time.sleep, hooks=[self.hook])

        ausleep(0.01)
        with self.auditor.start_auditing():
            ausleep(0.01)
        ausleep(0.01)

        recorded_time = self.hook.event_times["test.time.sleep"]
        self.assertAlmostEqual(recorded_time, 0.01, delta=0.005)

        # Check logging output
        self.auditor.log_results()
        self.logging_output.seek(0)
        output = [line.rstrip() for line in self.logging_output.readlines()]
        self.assertEqual(output[0], "(INFO) TimerHook results:")
        self.assertEqual(
            output[1], "(INFO)     Time spent in test.time.sleep: %f" % recorded_time
        )


if __name__ == "__main__":
    unittest.main()



# Tests for protocols and functions defined in seagrass.base

import seagrass.base as base
import typing as t
import unittest


class CustomHookImplementationTestCase(unittest.TestCase):
    def setUp(self):
        # Create a version of the ProtoHook protocol that we can check at runtime.
        self.CheckableProtoHook = t.runtime_checkable(base.ProtoHook)

    def test_get_prehook_and_posthook_priority(self):
        class MyHook:
            prehook_priority: int = 7

            def prehook(self, *args):
                ...

            def posthook(self, *args):
                ...

        hook = MyHook()
        self.assertIsInstance(
            hook,
            self.CheckableProtoHook,
            f"{hook.__class__.__name__} does not satisfy the hooking protocol",
        )
        self.assertEqual(base.prehook_priority(hook), 7)
        self.assertEqual(base.posthook_priority(hook), base.DEFAULT_POSTHOOK_PRIORITY)

        # The prehook_priority and posthook_priority are both required to be integers
        hook.prehook_priority = "Alice"
        with self.assertRaises(TypeError):
            base.prehook_priority(hook)

        hook.posthook_priority = None
        with self.assertRaises(TypeError):
            base.posthook_priority(hook)


if __name__ == "__main__":
    unittest.main()
